use std::sync::Arc;

use reqwest::Client;
use serde::{Deserialize, Serialize};
use tokio::io::AsyncBufReadExt;
use tokio::sync::RwLock;
use tokio_stream::wrappers::LinesStream;
use tokio_stream::{Stream, StreamExt};
use tokio_util::io::StreamReader;

pub struct OllamaAPI {
    pub api_url: String,
    client: Client,
    pub chats: Vec<Arc<RwLock<OllamaChat>>>,
}

#[derive(Serialize)]
pub struct OllamaChat {
    #[serde(skip_serializing)]
    pub history_size: u32,

    pub model: String,
    pub messages: Vec<ChatMessage>,
}

#[derive(Serialize, Deserialize, Clone)]
pub enum Role {
    #[serde(rename = "system")]
    System,
    #[serde(rename = "user")]
    User,
    #[serde(rename = "assistant")]
    Assistant,
}

#[derive(Serialize, Deserialize, Clone)]
pub struct ChatMessage {
    pub role: Role,
    pub content: String,
}

#[derive(Deserialize)]
struct OllamaResponse {
    message: ChatMessage,
}

fn deserialize_chunk(chunk: Result<String, std::io::Error>) -> anyhow::Result<String> {
    match chunk {
        Err(err) => Err(anyhow::Error::new(err)),
        Ok(str) => {
            let response = serde_json::from_str::<OllamaResponse>(&str)?;
            Ok(response.message.content)
        }
    }
}

impl OllamaAPI {
    pub async fn complete(
        &self,
        chat: &OllamaChat,
    ) -> anyhow::Result<impl Stream<Item = anyhow::Result<String>>> {
        let body = serde_json::to_string(&chat)?;
        println!("Sending: {}", body);

        let request = self
            .client
            .post(self.api_url.clone() + "/api/chat")
            .body(serde_json::to_string(&chat)?)
            .build()?;
        let stream = self
            .client
            .execute(request)
            .await?
            .bytes_stream()
            .map(|x| match x {
                Err(err) => Err(std::io::Error::new(std::io::ErrorKind::Other, err)),
                Ok(bytes) => Ok(bytes),
            });
        let lines = StreamReader::new(stream).lines();
        Ok(LinesStream::new(lines).map(deserialize_chunk))
    }

    pub fn create(api_url: &str) -> anyhow::Result<Self> {
        let client = Client::builder().build()?;
        Ok(Self {
            api_url: String::from(api_url),
            client,
            chats: Vec::new(),
        })
    }

    pub fn create_chat(&mut self, model: &str) -> Arc<RwLock<OllamaChat>> {
        let chat = Arc::new(RwLock::new(OllamaChat {
            history_size: 0,
            model: String::from(model),
            messages: Vec::new(),
        }));

        self.chats.push(chat.clone());
        chat
    }
}

impl OllamaChat {
    pub fn send_message(&mut self, role: Role, content: &str) {
        self.messages.push(ChatMessage {
            role,
            content: String::from(content),
        })
    }

    pub fn send_system(&mut self, content: &str) {
        self.send_message(Role::System, content);
    }

    pub fn send_user(&mut self, content: &str) {
        self.send_message(Role::User, content);
    }

    pub fn send_assistant(&mut self, content: &str) {
        self.send_message(Role::Assistant, content);
    }
}
