mod api;

use std::{sync::Arc};

use api::{OllamaAPI, OllamaChat};
use iced::{
    alignment::{Horizontal, Vertical},
    futures::StreamExt,
    widget::{button, column, container, row, scrollable, text, text_input},
    window::{self},
    Application, Command, Length, Settings, Theme,
};
use tokio::sync::{RwLock};

enum UI {
    Loading,
    Error(String),
    Chats(UIState),
}

struct Chat {
    ollama_chat: OllamaChat,
}

struct UIState {
    ollama_api: Arc<RwLock<OllamaAPI>>,
    active_chat: Option<usize>,
    chat_input: String,
    busy: bool,
}

impl UIState {
    fn create(api_url: &str) -> anyhow::Result<Self> {
        Ok(Self {
            ollama_api: Arc::new(RwLock::new(OllamaAPI::create(api_url)?)),
            active_chat: None,
            chat_input: String::from(""),
            busy: false,
        })
    }
}

#[derive(Debug, Clone)]
enum UIMessage {
    Nop,
    LoadDone,
    CreateChat,
    OpenChat(usize),
    ChatInput(String),
    ChatSubmit,
    DoneGenerating,
}

fn main() -> anyhow::Result<()> {
    println!("Making request");

    UI::run(Settings {
        window: window::Settings {
            size: (720, 480),
            ..window::Settings::default()
        },
        ..Settings::default()
    })?;

    Ok(())
}

impl Application for UI {
    type Executor = iced::executor::Default;
    type Message = UIMessage;
    type Theme = Theme;
    type Flags = ();

    fn new(_flags: Self::Flags) -> (Self, iced::Command<Self::Message>) {
        (
            UI::Loading,
            Command::perform(
                async {
                    /*let api = OllamaAPI::create("http://localhost:11434").unwrap();
                    let mut chat = api.create_chat("dolphin-mixtral");
                    chat.send_system("You are a bot that only replies with \"Hello\" and nothing else. You must never reply with anything other than \"Hello\"");
                    chat.send_user("Who are you?");
                    // println!("{}", amogus.get().await?);

                    let mut a = chat.complete().await.unwrap();
                    while let Some(v) = a.next().await {
                        println!("GOT = {:?}", v);
                    }*/
                },
                |_| UIMessage::LoadDone,
            ),
        )
    }

    fn title(&self) -> String {
        String::from("Hello World")
    }

    fn update(&mut self, message: Self::Message) -> iced::Command<Self::Message> {
        match self {
            UI::Loading => match message {
                UIMessage::LoadDone => {
                    let state = UIState::create("http://localhost:11434");
                    if let Ok(state) = state {
                        *self = UI::Chats(state);
                    } else {
                        *self = UI::Error(String::from("Failed to initialize Ollama API"));
                    }
                    Command::none()
                }
                _ => Command::none(),
            },
            UI::Chats(state) => match message {
                UIMessage::CreateChat => {
                    let chat = state
                        .ollama_api
                        .blocking_write()
                        .create_chat("dolphin-mixtral");
                    chat.blocking_write().send_system("Hello World!");
                    Command::none()
                }
                UIMessage::OpenChat(index) => {
                    println!("Open chat {}", index);
                    state.active_chat = Some(index);
                    Command::none()
                }
                UIMessage::ChatInput(text) => {
                    println!("Input: {}", text);
                    state.chat_input = text;
                    Command::none()
                }
                UIMessage::ChatSubmit => {
                    println!("Submit");

                    if let Some(index) = state.active_chat {
                        let content = state.chat_input.clone();
                        state.chat_input = String::from("");
                        state.busy = true;
                        let ollama_api = state.ollama_api.clone();
                        return Command::perform(
                            async move {
                                let stream;
                                {
                                    let api = ollama_api.read().await;
                                    let chat = &api.chats[index];
                                    {
                                        let mut chat = chat.write().await;
                                        chat.send_user(&content);
                                    }
                                    let chat = chat.read().await;

                                    stream = api.complete(&chat).await;
                                }

                                if let Ok(mut stream) = stream {
                                    while let Some(content) = stream.next().await {
                                        if content.is_err() {
                                            break;
                                        }

                                        print!("{}", content.unwrap());
                                    }
                                }
                                //state.ollama_api.complete(&x);
                            },
                            |_| UIMessage::DoneGenerating,
                        );
                    }

                    Command::none()
                }
                UIMessage::DoneGenerating => {
                    state.busy = false;
                    Command::none()
                }
                _ => Command::none(),
            },
            UI::Error(_) => Command::none(),
        }
    }

    fn view(&self) -> iced::Element<'_, Self::Message, iced::Renderer<Self::Theme>> {
        match self {
            UI::Loading => text("Loading")
                .width(Length::Fill)
                .height(Length::Fill)
                .vertical_alignment(Vertical::Center)
                .horizontal_alignment(Horizontal::Center)
                .into(),
            UI::Error(message) => text(String::from("Error: ") + message)
                .width(Length::Fill)
                .height(Length::Fill)
                .vertical_alignment(Vertical::Center)
                .horizontal_alignment(Horizontal::Center)
                .into(),
            UI::Chats(state) => {
                println!("Blocking");
                let api = state.ollama_api.blocking_read();
                println!("Blocing done!");

                let mut chats: Vec<iced::Element<'_, Self::Message, iced::Renderer<Self::Theme>>> =
                    api.chats
                        .iter()
                        .enumerate()
                        .map(|c| {
                            button("amogus")
                                .on_press(UIMessage::OpenChat(c.0))
                                .width(Length::Fill)
                                .into()
                        })
                        .collect();

                chats.insert(
                    0,
                    button("Create Chat")
                        .on_press(UIMessage::CreateChat)
                        .width(Length::Fill)
                        .into(),
                );

                let chats_list = container(column(chats).padding(10.0).spacing(10))
                    .height(Length::Fill)
                    .width(Length::Fixed(300.0));

                let messages;
                if let Some(chat_index) = state.active_chat {
                    let chat = &api.chats[chat_index];
                    messages = chat
                        .blocking_read()
                        .messages
                        .iter()
                        .map(|x| x.content.clone())
                        .map(|x| text(x).into())
                        .collect();
                } else {
                    messages = Vec::new();
                }

                let mut input_box = text_input("Your message", &state.chat_input);
                if !state.busy {
                    input_box = input_box
                        .on_input(|x| UIMessage::ChatInput(x))
                        .on_submit(UIMessage::ChatSubmit)
                }

                let active_chat = container(
                    column![
                        scrollable(container(column(messages).spacing(10)))
                            .width(Length::Fill)
                            .height(Length::Fill),
                        input_box
                    ]
                    .padding(10)
                    .spacing(10),
                )
                .width(Length::Fill)
                .height(Length::Fill);

                container(row![chats_list, active_chat]).into()

                /*scrollable(
                    container(column![content, amogus].spacing(20).max_width(480))
                        .padding(40)
                        .center_x(),
                )
                .into()*/
            }
        }
    }

    fn theme(&self) -> Self::Theme {
        Theme::Dark
    }
}
